@@ -47,6 +47,7 @@
 
 /** window title */
 static char *title = "EventDemo";
+static char *appid = "org.freedesktop.weston.eventdemo";
 
 /** window width */
 static int width = 500;
@@ -55,7 +56,7 @@ static int width = 500;
 static int height = 400;
 
 /** set if window has no borders */
-static int noborder = 0;
+static bool noborder = false;
 
 /** if non-zero, maximum window width */
 static int width_max = 0;
@@ -64,25 +65,25 @@ static int width_max = 0;
 static int height_max = 0;
 
 /** set to log redrawing */
-static int log_redraw = 0;
+static bool log_redraw = false;
 
 /** set to log resizing */
-static int log_resize = 0;
+static bool log_resize = false;
 
 /** set to log keyboard focus */
-static int log_focus = 0;
+static bool log_focus = false;
 
 /** set to log key events */
-static int log_key = 0;
+static bool log_key = false;
 
 /** set to log button events */
-static int log_button = 0;
+static bool log_button = false;
 
 /** set to log axis events */
-static int log_axis = 0;
+static bool log_axis = false;
 
 /** set to log motion events */
-static int log_motion = 0;
+static bool log_motion = false;
 
 /**
  * \struct eventdemo
@@ -350,8 +351,6 @@ axis_discrete_handler(struct widget *widget, struct input *input,
  * \param widget widget
  * \param input input device that caused the motion event
  * \param time time the event happened
- * \param x absolute x position
- * \param y absolute y position
  * \param x x position relative to the window
  * \param y y position relative to the window
  * \param data user data associated to the window
@@ -400,6 +399,7 @@ eventdemo_create(struct display *d)
 	} else {
 		e->widget = window_frame_create(e->window, e);
 		window_set_title(e->window, title);
+		window_set_appid(e->window, appid);
 	}
 	e->display = d;
 
@@ -513,7 +513,7 @@ main(int argc, char *argv[])
 	if (!log_redraw && !log_resize && !log_focus && !log_key &&
 	    !log_button && !log_axis && !log_motion)
 	  log_redraw = log_resize = log_focus = log_key =
-	    log_button = log_axis = log_motion = 1;
+	    log_button = log_axis = log_motion = true;
 
 	/* Connect to the display and have the arguments parsed */
 	d = display_create(&argc, argv);
