@@ -26,11 +26,13 @@
 #include "config.h"
 
 #include <assert.h>
+#include <ctype.h>
 #include <fcntl.h>
 #include <unistd.h>
 #include <stdio.h>
 #include <libudev.h>
 #include <sys/mman.h>
+#include <time.h>
 
 #include "shared/helpers.h"
 #include "shared/platform.h"
@@ -39,6 +41,7 @@
 #include <libweston/pixel-formats.h>
 #include "xdg-shell-client-protocol.h"
 #include "linux-dmabuf-unstable-v1-client-protocol.h"
+#include "presentation-time-client-protocol.h"
 
 #include <xf86drm.h>
 #include <gbm.h>
@@ -47,7 +50,12 @@
 #include <GLES2/gl2.h>
 #include <GLES2/gl2ext.h>
 
-#define NUM_BUFFERS 3
+#define L_LINE "│   "
+#define L_VAL  "├───"
+#define L_LAST "└───"
+#define L_GAP  "    "
+
+#define NUM_BUFFERS 4
 
 /* We have to hack the DRM-backend to pretend that planes of the underlying
  * hardware don't support this format. If you change the value of this constant,
@@ -60,7 +68,10 @@ static const char *vert_shader_text =
 	"attribute vec4 color;\n"
 	"varying vec4 v_color;\n"
 	"void main() {\n"
-	"	gl_Position = pos;\n"
+	"	// We need to render upside-down, because rendering through an\n"
+	"	// FBO causes the bottom of the image to be written to the top\n"
+	"	// pixel row of the buffer, y-flipping the image.\n"
+	"	gl_Position = vec4(1.0, -1.0, 1.0, 1.0) * pos;\n"
 	"	v_color = color;\n"
 	"}\n";
 
@@ -108,6 +119,9 @@ struct output {
 	int width, height;
 	int scale;
 	bool initialized;
+	struct {
+		int width, height;
+	} configure;
 };
 
 struct egl {
@@ -133,15 +147,21 @@ struct display {
 	struct output output;
 	struct xdg_wm_base *wm_base;
 	struct zwp_linux_dmabuf_v1 *dmabuf;
+	struct wp_presentation *presentation;
 	struct gbm_device *gbm_device;
 	struct egl egl;
 };
 
 struct buffer {
+	bool created;
+	bool valid;
 	struct window *window;
 	struct wl_buffer *buffer;
-	bool busy;
-	bool recreate;
+	enum {
+		NOT_CREATED,
+		IN_USE,
+		AVAILABLE
+	} status;
 	int dmabuf_fds[4];
 	struct gbm_bo *bo;
 	EGLImageKHR egl_image;
@@ -160,12 +180,14 @@ struct window {
 	struct xdg_surface *xdg_surface;
 	struct xdg_toplevel *xdg_toplevel;
 	struct wl_callback *callback;
+	struct wp_presentation_feedback *presentation_feedback;
 	bool wait_for_configure;
-	uint32_t n_redraws;
+	bool presented_zero_copy;
 	struct zwp_linux_dmabuf_feedback_v1 *dmabuf_feedback_obj;
 	struct dmabuf_feedback dmabuf_feedback, pending_dmabuf_feedback;
 	int card_fd;
 	struct drm_format format;
+	uint32_t bo_flags;
 	struct buffer buffers[NUM_BUFFERS];
 };
 
@@ -442,26 +464,30 @@ buffer_free(struct buffer *buf)
 
 	for (i = 0; i < buf->num_planes; i++)
 		close(buf->dmabuf_fds[i]);
+
+	buf->created = false;
 }
 
 static void
 create_dmabuf_buffer(struct window *window, struct buffer *buf, uint32_t width,
 		     uint32_t height, uint32_t format, unsigned int count_modifiers,
-		     uint64_t *modifiers);
+		     uint64_t *modifiers, uint32_t bo_flags);
 
 static void
-buffer_recreate(struct buffer *buf)
+buffer_recreate(struct buffer *buf, struct window *window)
 {
-	struct window *window = buf->window;
-	uint32_t width = buf->width;
-	uint32_t height = buf->height;
+	uint32_t width = window->display->output.width;
+	uint32_t height = window->display->output.height;
+
+	if (buf->created)
+		buffer_free(buf);
 
-	buffer_free(buf);
 	create_dmabuf_buffer(window, buf, width, height,
 			     window->format.format,
 			     window->format.modifiers.size / sizeof(uint64_t),
-			     window->format.modifiers.data);
-	buf->recreate = false;
+			     window->format.modifiers.data, window->bo_flags);
+	buf->created = true;
+	buf->valid = true;
 }
 
 static void
@@ -469,10 +495,7 @@ buffer_release(void *data, struct wl_buffer *buffer)
 {
 	struct buffer *buf = data;
 
-	buf->busy = false;
-
-	if (buf->recreate)
-		buffer_recreate(buf);
+	buf->status = AVAILABLE;
 }
 
 static const struct wl_buffer_listener buffer_listener = {
@@ -485,6 +508,7 @@ create_succeeded(void *data, struct zwp_linux_buffer_params_v1 *params,
 {
 	struct buffer *buf = data;
 
+	buf->status = AVAILABLE;
 	buf->buffer = new_buffer;
 	wl_buffer_add_listener(buf->buffer, &buffer_listener, buf);
 	zwp_linux_buffer_params_v1_destroy(params);
@@ -509,33 +533,40 @@ static const struct zwp_linux_buffer_params_v1_listener params_listener = {
 static void
 create_dmabuf_buffer(struct window *window, struct buffer *buf, uint32_t width,
 		     uint32_t height, uint32_t format, unsigned int count_modifiers,
-		     uint64_t *modifiers)
+		     uint64_t *modifiers, uint32_t bo_flags)
 {
 	struct display *display = window->display;
 	static uint32_t flags = 0;
 	struct zwp_linux_buffer_params_v1 *params;
 	int i;
 
+	buf->status = NOT_CREATED;
 	buf->window = window;
 	buf->width = width;
 	buf->height = height;
 	buf->format = format;
 
-#ifdef HAVE_GBM_MODIFIERS
 	if (count_modifiers > 0) {
+#ifdef HAVE_GBM_BO_CREATE_WITH_MODIFIERS2
+		buf->bo = gbm_bo_create_with_modifiers2(display->gbm_device,
+							buf->width, buf->height,
+							format, modifiers,
+							count_modifiers,
+							bo_flags);
+#else
 		buf->bo = gbm_bo_create_with_modifiers(display->gbm_device,
 						       buf->width, buf->height,
 						       format, modifiers,
 						       count_modifiers);
+#endif
 		if (buf->bo)
 			buf->modifier = gbm_bo_get_modifier(buf->bo);
 	}
-#endif
 
 	if (!buf->bo) {
 		buf->bo = gbm_bo_create(display->gbm_device, buf->width,
 					buf->height, buf->format,
-					GBM_BO_USE_RENDERING);
+					bo_flags);
 		buf->modifier = DRM_FORMAT_MOD_INVALID;
 	}
 
@@ -572,9 +603,30 @@ window_next_buffer(struct window *window)
 {
 	unsigned int i;
 
+	for (i = 0; i < NUM_BUFFERS; i++) {
+		if (!window->buffers[i].created ||
+		    (!window->buffers[i].valid &&
+		     window->buffers[i].status == AVAILABLE))
+			buffer_recreate(&window->buffers[i], window);
+	}
+
 	for (i = 0; i < NUM_BUFFERS; i++)
-		if (!window->buffers[i].busy)
+		if (window->buffers[i].status == AVAILABLE)
 			return &window->buffers[i];
+
+	while (true) {
+		/* In this client, we create buffers lazily and also sometimes
+		* have to recreate the buffers. As we are not using the
+		* create_immed request from zwp_linux_dmabuf_v1, we need to wait
+		* for an event from the server (what leads to create_succeeded()
+		* being called in this client). */
+		wl_display_roundtrip(window->display->display);
+
+		for (i = 0; i < NUM_BUFFERS; i++)
+			if (window->buffers[i].status == AVAILABLE)
+				return &window->buffers[i];
+	}
+
 	return NULL;
 }
 
@@ -613,10 +665,11 @@ render(struct buffer *buffer)
 	glDisableVertexAttribArray(window->gl.pos);
 	glDisableVertexAttribArray(window->gl.color);
 
-	glFinish();
+	glFlush();
 }
 
 static const struct wl_callback_listener frame_listener;
+static const struct wp_presentation_feedback_listener presentation_feedback_listener;
 
 static void
 redraw(void *data, struct wl_callback *callback, uint32_t time)
@@ -638,27 +691,103 @@ redraw(void *data, struct wl_callback *callback, uint32_t time)
 
 	window->callback = wl_surface_frame(window->surface);
 	wl_callback_add_listener(window->callback, &frame_listener, window);
+
+	if (window->presentation_feedback) {
+		wp_presentation_feedback_destroy(window->presentation_feedback);
+		window->presentation_feedback = NULL;
+	}
+	if (window->display->presentation) {
+		window->presentation_feedback =
+			wp_presentation_feedback(window->display->presentation,
+						 window->surface);
+		wp_presentation_feedback_add_listener(window->presentation_feedback,
+						      &presentation_feedback_listener,
+						      window);
+	}
+
 	wl_surface_commit(window->surface);
-	buf->busy = true;
+	buf->status = IN_USE;
 
 	region = wl_compositor_create_region(window->display->compositor);
 	wl_region_add(region, 0, 0, window->display->output.width,
 		      window->display->output.height);
 	wl_surface_set_opaque_region(window->surface, region);
 	wl_region_destroy(region);
-
-	window->n_redraws++;
 }
 
 static const struct wl_callback_listener frame_listener = {
 	redraw
 };
 
+static void presentation_feedback_handle_sync_output(void *data,
+						     struct wp_presentation_feedback *feedback,
+						     struct wl_output *output)
+{
+}
+
+static void presentation_feedback_handle_presented(void *data,
+						   struct wp_presentation_feedback *feedback,
+						   uint32_t tv_sec_hi,
+						   uint32_t tv_sec_lo,
+						   uint32_t tv_nsec,
+						   uint32_t refresh,
+						   uint32_t seq_hi,
+						   uint32_t seq_lo,
+						   uint32_t flags)
+{
+	struct window *window = data;
+	bool zero_copy = flags & WP_PRESENTATION_FEEDBACK_KIND_ZERO_COPY;
+
+	if (zero_copy && !window->presented_zero_copy) {
+		fprintf(stderr, "Presenting in zero-copy mode\n");
+	}
+	if (!zero_copy && window->presented_zero_copy) {
+		fprintf(stderr, "Stopped presenting in zero-copy mode\n");
+	}
+
+	window->presented_zero_copy = zero_copy;
+	wp_presentation_feedback_destroy(feedback);
+	window->presentation_feedback = NULL;
+}
+
+static void presentation_feedback_handle_discarded(void *data,
+						   struct wp_presentation_feedback *feedback)
+{
+	struct window *window = data;
+	wp_presentation_feedback_destroy(feedback);
+	window->presentation_feedback = NULL;
+}
+
+static const struct wp_presentation_feedback_listener presentation_feedback_listener = {
+	.sync_output = presentation_feedback_handle_sync_output,
+	.presented = presentation_feedback_handle_presented,
+	.discarded = presentation_feedback_handle_discarded,
+};
+
+
+static void
+window_buffers_invalidate(struct window *window)
+{
+	unsigned int i;
+
+	for (i = 0; i < NUM_BUFFERS; i++)
+		window->buffers[i].valid = false;
+}
+
 static void
 xdg_surface_handle_configure(void *data, struct xdg_surface *surface,
 			     uint32_t serial)
 {
 	struct window *window = data;
+	struct output *output = &window->display->output;
+
+	if (output->configure.width != output->width ||
+	    output->configure.height != output->height) {
+		output->width = output->configure.width;
+		output->height = output->configure.height;
+
+		window_buffers_invalidate (window);
+	}
 
 	xdg_surface_ack_configure(surface, serial);
 	window->wait_for_configure = false;
@@ -673,6 +802,11 @@ xdg_toplevel_handle_configure(void *data, struct xdg_toplevel *toplevel,
 			      int32_t width, int32_t height,
 			      struct wl_array *states)
 {
+	struct window *window = data;
+	struct output *output = &window->display->output;
+
+	output->configure.width = width;
+	output->configure.height = height;
 }
 
 static void
@@ -714,7 +848,7 @@ egl_setup(struct window *window)
 	assert(egl->display && "error: could not create EGL display");
 
 	ret = eglInitialize(egl->display, &major, &minor);
-	assert(ret != EGL_FALSE && "error: failed to intialized EGL display");
+	assert(ret != EGL_FALSE && "error: failed to initialized EGL display");
 
 	ret = eglBindAPI(EGL_OPENGL_ES_API);
 	assert(ret != EGL_FALSE && "error: failed to set EGL API");
@@ -790,6 +924,8 @@ destroy_window(struct window *window)
 
 	if (window->callback)
 		wl_callback_destroy(window->callback);
+	if (window->presentation_feedback)
+		wp_presentation_feedback_destroy(window->presentation_feedback);
 
 	for (i = 0; i < NUM_BUFFERS; i++)
 		if (window->buffers[i].buffer)
@@ -818,9 +954,6 @@ static struct window *
 create_window(struct display *display)
 {
 	struct window *window;
-	uint32_t width = display->output.width;
-	uint32_t height	= display->output.height;
-	unsigned int i;
 
 	window = zalloc(sizeof *window);
 	assert(window && "error: failed to allocate memory for window");
@@ -849,12 +982,6 @@ create_window(struct display *display)
 	egl_setup(window);
 	gl_setup(window);
 
-	for (i = 0; i < NUM_BUFFERS; i++)
-		create_dmabuf_buffer(window, &window->buffers[i], width, height,
-				     window->format.format,
-				     window->format.modifiers.size / sizeof(uint64_t),
-				     window->format.modifiers.data);
-
 
 	window->xdg_surface = xdg_wm_base_get_xdg_surface(display->wm_base,
 							  window->surface);
@@ -870,14 +997,13 @@ create_window(struct display *display)
 	xdg_toplevel_set_title(window->xdg_toplevel, "simple-dmabuf-feedback");
 	xdg_toplevel_set_app_id(window->xdg_toplevel,
 			"org.freedesktop.weston.simple-dmabuf-feedback");
+	xdg_toplevel_set_fullscreen(window->xdg_toplevel, NULL);
 
 	window->wait_for_configure = true;
 	wl_surface_commit(window->surface);
 
 	wl_display_roundtrip(display->display);
 
-	xdg_toplevel_set_fullscreen(window->xdg_toplevel, NULL);
-
 	assert(!window->wait_for_configure &&
 	       "error: could not configure XDG surface");
 
@@ -980,8 +1106,7 @@ dmabuf_feedback_main_device(void *data,
 	drm_node = get_drm_node(feedback->main_device, false);
 	assert(drm_node && "error: failed to retrieve drm node");
 
-	fprintf(stderr, "compositor sent main_device event for dma-buf feedback - %s\n",
-		drm_node);
+	fprintf(stderr, "feedback: main device %s\n", drm_node);
 
 	if (!window->card_fd) {
 		window->card_fd = open(drm_node, O_RDWR | O_CLOEXEC);
@@ -1050,18 +1175,54 @@ dmabuf_feedback_tranche_formats(void *data,
 	}
 }
 
+static char
+bits2graph(uint32_t value, unsigned bitoffset)
+{
+	int c = (value >> bitoffset) & 0xff;
+
+	if (isgraph(c) || isspace(c))
+		return c;
+
+	return '?';
+}
+
+static void
+fourcc2str(uint32_t format, char *str, int len)
+{
+	int i;
+
+	assert(len >= 5);
+
+	for (i = 0; i < 4; i++)
+		str[i] = bits2graph(format, i * 8);
+	str[i] = '\0';
+}
+
 static void
 print_tranche_format_modifier(uint32_t format, uint64_t modifier)
 {
 	const struct pixel_format_info *fmt_info;
+	char *format_str;
 	char *mod_name;
+	int len;
 
-	fmt_info = pixel_format_get_info(format);
 	mod_name = pixel_format_get_modifier(modifier);
+	fmt_info = pixel_format_get_info(format);
 
-	fprintf(stderr, "│	├────────tranche format/modifier pair - format %s, modifier %s\n",
-			fmt_info ? fmt_info->drm_format_name : "UNKNOWN", mod_name);
+	if (fmt_info) {
+		len = asprintf(&format_str, "%s", fmt_info->drm_format_name);
+	} else {
+		char fourcc_str[5];
 
+		fourcc2str(format, fourcc_str, sizeof(fourcc_str));
+		len = asprintf(&format_str, "%s (0x%08x)", fourcc_str, format);
+	}
+	assert(len > 0);
+
+	fprintf(stderr, L_LINE L_VAL " format %s, modifier %s\n",
+		format_str, mod_name);
+
+	free(format_str);
 	free(mod_name);
 }
 
@@ -1075,14 +1236,14 @@ print_dmabuf_feedback_tranche(struct dmabuf_feedback_tranche *tranche)
 	drm_node = get_drm_node(tranche->target_device, tranche->is_scanout_tranche);
 	assert(drm_node && "error: could not retrieve drm node");
 
-	fprintf(stderr, "├──────target_device for tranche - %s\n", drm_node);
-	fprintf(stderr, "│	└scanout tranche? %s\n", tranche->is_scanout_tranche ? "yes" : "no");
+	fprintf(stderr, L_VAL " tranche: target device %s, %s\n",
+		drm_node, tranche->is_scanout_tranche ? "scanout" : "no flags");
 
 	wl_array_for_each(fmt, &tranche->formats.arr)
 		wl_array_for_each(mod, &fmt->modifiers)
 			print_tranche_format_modifier(fmt->format, *mod);
 
-	fprintf(stderr, "│	└end of tranche\n");
+	fprintf(stderr, L_LINE L_LAST " end of tranche\n");
 }
 
 static void
@@ -1103,7 +1264,7 @@ dmabuf_feedback_tranche_done(void *data,
 	dmabuf_feedback_tranche_init(&feedback->pending_tranche);
 }
 
-static void
+static bool
 pick_initial_format_from_renderer_tranche(struct window *window,
 					  struct dmabuf_feedback_tranche *tranche)
 {
@@ -1117,13 +1278,14 @@ pick_initial_format_from_renderer_tranche(struct window *window,
 		window->format.format = fmt->format;
 		wl_array_copy(&window->format.modifiers, &fmt->modifiers);
 
-		return;
-	}
+		window->bo_flags = GBM_BO_USE_RENDERING;
 
-	assert(0 && "error: INITIAL_BUFFER_FORMAT not supported by the hardware");
+		return true;
+	}
+	return false;
 }
 
-static void
+static bool
 pick_format_from_scanout_tranche(struct window *window,
 				 struct dmabuf_feedback_tranche *tranche)
 {
@@ -1132,8 +1294,9 @@ pick_format_from_scanout_tranche(struct window *window,
 
 	wl_array_for_each(fmt, &tranche->formats.arr) {
 
-		/* Ignore format that we're already using. */
-		if (fmt->format == window->format.format)
+		/* Ignore the format that we want to pick from the render
+		 * tranche. */
+		if (fmt->format == INITIAL_BUFFER_FORMAT)
 			continue;
 
 		/* Format should be supported by the compositor. */
@@ -1147,10 +1310,11 @@ pick_format_from_scanout_tranche(struct window *window,
 		window->format.format = fmt->format;
 		wl_array_copy(&window->format.modifiers, &fmt->modifiers);
 
-		return;
-	}
+		window->bo_flags = GBM_BO_USE_RENDERING | GBM_BO_USE_SCANOUT;
 
-	assert(0 && "error: no valid pair of format/modifier in the scanout tranche");
+		return true;
+	}
+	return false;
 }
 
 static void
@@ -1158,25 +1322,35 @@ dmabuf_feedback_done(void *data, struct zwp_linux_dmabuf_feedback_v1 *dmabuf_fee
 {
 	struct window *window = data;
 	struct dmabuf_feedback_tranche *tranche;
-	unsigned int i;
+	bool got_scanout_tranche = false;
 
-	fprintf(stderr, "└end of dma-buf feedback\n\n");
+	fprintf(stderr, L_LAST " end of dma-buf feedback\n\n");
 
 	/* The first time that we receive dma-buf feedback for a surface it
-	 * contains only the renderer tranche. We pick the INITIAL_BUFFER_FORMAT
+	 * contains only the renderer tranches. We pick the INITIAL_BUFFER_FORMAT
 	 * from there. Then the compositor should detect that the format is
 	 * unsupported by the underlying hardware (not actually, but you should
-	 * have faked this in the DRM-backend) and send the scanout tranche. We
-	 * use the formats/modifiers of the scanout tranche to reallocate our
+	 * have faked this in the DRM-backend) and send the scanout tranches. We
+	 * use the formats/modifiers of the scanout tranches to reallocate our
 	 * buffers. */
 	wl_array_for_each(tranche, &window->pending_dmabuf_feedback.tranches) {
 		if (tranche->is_scanout_tranche) {
-			pick_format_from_scanout_tranche(window, tranche);
-			for (i = 0; i < NUM_BUFFERS; i++)
-				window->buffers[i].recreate = true;
-			break;
+			got_scanout_tranche = true;
+			if (pick_format_from_scanout_tranche(window, tranche)) {
+				window_buffers_invalidate(window);
+				break;
+			}
 		}
-		pick_initial_format_from_renderer_tranche(window, tranche);
+		if (pick_initial_format_from_renderer_tranche(window, tranche))
+			break;
+	}
+
+	if (got_scanout_tranche) {
+		assert(window->format.format != INITIAL_BUFFER_FORMAT &&
+		       "error: no valid pair of format/modifier in the scanout tranches");
+	} else {
+		assert(window->format.format == INITIAL_BUFFER_FORMAT &&
+		       "error: INITIAL_BUFFER_FORMAT not supported by the hardware");
 	}
 
 	dmabuf_feedback_fini(&window->dmabuf_feedback);
@@ -1209,12 +1383,6 @@ static void
 output_handle_mode(void *data, struct wl_output *wl_output, uint32_t flags,
 		   int width, int height, int refresh)
 {
-	struct output *output = data;
-
-	if (flags & WL_OUTPUT_MODE_CURRENT) {
-		output->width = width;
-		output->height = height;
-	}
 }
 
 static void
@@ -1277,6 +1445,10 @@ registry_handle_global(void *data, struct wl_registry *registry,
 		d->dmabuf = wl_registry_bind(registry, id,
 					     &zwp_linux_dmabuf_v1_interface,
 					     MIN(version, 4));
+	} else if (strcmp(interface, "wp_presentation") == 0) {
+		d->presentation = wl_registry_bind(registry, id,
+					           &wp_presentation_interface,
+						   1);
 	}
 }
 
@@ -1301,6 +1473,9 @@ destroy_display(struct display *display)
 	if (display->egl.display != EGL_NO_DISPLAY)
 		eglTerminate(display->egl.display);
 
+	if (display->presentation)
+		wp_presentation_destroy(display->presentation);
+
 	zwp_linux_dmabuf_v1_destroy(display->dmabuf);
 	xdg_wm_base_destroy(display->wm_base);
 
@@ -1340,49 +1515,40 @@ create_display()
 }
 
 /* Simple client to test the dma-buf feedback implementation. This does not
- * replace the need to implement a dma-buf feedback test that can be run in
- * the CI. But as we still don't know exactly how to do this, this client
- * can be helpful to run tests manually.
+ * replace the need to implement a dma-buf feedback test that can be run in the
+ * CI. But as we still don't know exactly how to do this, this client can be
+ * helpful to run tests manually. It can also be helpful to test other
+ * compositors.
  *
- * In order to use this, we have to hack the DRM-backend to pretend that
- * INITIAL_BUFFER_FORMAT is not supported by the planes of the underlying
- * hardware. In Weston, we have to do this in
- * drm_output_prepare_plane_view(), more specifically in the part where
- * we call drm_output_plane_view_has_valid_format(). So we'd have something
- * like this:
+ * In order to use this client, we have to hack the DRM-backend of the
+ * compositor to pretend that INITIAL_BUFFER_FORMAT is not supported by the
+ * planes of the underlying hardware.
  *
- *     // in this example, INITIAL_BUFFER_FORMAT == DRM_FORMAT_XRGB8888
+ * How this client works:
  *
- *     bool fake_unsupported_format = false;
- *     if (fb && fb->format->format == DRM_FORMAT_XRGB8888)
- *             fake_unsupported_format = true;
+ * This client creates a surface and buffers for it with the same resolution of
+ * the output mode in use. Also, it sets the surface to fullscreen. So we have
+ * everything set to allow the surface to be placed in an overlay plane. But as
+ * these buffers are created with INITIAL_BUFFER_FORMAT, they are not eligible
+ * for direct scanout.
  *
- *     if (!drm_output_plane_view_has_valid_format(plane, state, ev, fb) ||
- *         fake_unsupported_format)
- *         ...
- *
- * It creates a surface and buffers for it with the same resolution of the
- * output mode in use. Also, it sets the surface to fullscreen. So we have
- * everything set to allow the surface to be placed in a plane. But as
- * these buffers are created with INITIAL_BUFFER_FORMAT, they are placed in
- * the renderer.
- *
- * When the compositor creates the client surface, it adds only the
- * renderer tranche to its dma-buf feedback object and send the feedback to
- * the client. But as the repaint cycles start and Weston detects that the
- * only reason why the surface has not been placed in a plane was the
- * incompatibility between the framebuffer format and the ones supported by
- * the planes of the underlying hardware, Weston adds a scanout tranche to
- * the surface dma-buf feedback and resend them. In this tranche the client
- * can find pairs of formats and modifiers supported by the planes, and so
- * it can recreate its buffers using one of these pairs in order to
- * increase the chances of its surface end up in a plane. */
+ * When Weston creates a client's surface, it adds only the renderer tranche to
+ * its dma-buf feedback object and send the feedback to the client. But as the
+ * repaint cycles start and Weston detects that the view of the client is not
+ * eligible for direct scanout because of the incompatibility of the
+ * framebuffer's format/modifier pair and the KMS device, it adds a scanout
+ * tranche to the feedback and resend it. In this scanout tranche the client can
+ * find parameters to re-allocate its buffers and increase its chances of
+ * hitting direct scanout. */
 int
 main(int argc, char **argv)
 {
 	struct display *display;
 	struct window *window;
 	int ret = 0;
+	struct timespec start_time, current_time;
+	const time_t MAX_TIME_SECONDS = 3;
+	time_t delta_time = 0;
 
 	fprintf(stderr, "This client was written with the purpose of manually test " \
 			"Weston's dma-buf feedback implementation. See main() " \
@@ -1391,9 +1557,14 @@ main(int argc, char **argv)
 	display = create_display();
 	window = create_window(display);
 
+	clock_gettime(CLOCK_MONOTONIC, &start_time);
+
 	redraw(window, NULL, 0);
-	while (ret != -1 && window->n_redraws < 200)
+	while (ret != -1 && delta_time < MAX_TIME_SECONDS) {
 		ret = wl_display_dispatch(display->display);
+		clock_gettime(CLOCK_MONOTONIC, &current_time);
+		delta_time = current_time.tv_sec - start_time.tv_sec;
+	}
 
 	destroy_window(window);
 	destroy_display(display);
