@@ -48,7 +48,6 @@
 #include <wayland-cursor.h>
 #include <libweston/zalloc.h>
 #include "xdg-shell-client-protocol.h"
-#include "fullscreen-shell-unstable-v1-client-protocol.h"
 #include "linux-dmabuf-unstable-v1-client-protocol.h"
 #include "weston-direct-display-client-protocol.h"
 #include "viewporter-client-protocol.h"
@@ -89,7 +88,7 @@ static inline const char *
 dump_format(uint32_t format, char out[4])
 {
 #if BYTE_ORDER == BIG_ENDIAN
-	format = __builtin_bswap32(format);
+	format = bswap32(format);
 #endif
 	memcpy(out, &format, 4);
 	return out;
@@ -117,7 +116,6 @@ struct display {
 	struct wl_cursor *default_cursor;
 	struct wl_surface *cursor_surface;
 	struct xdg_wm_base *wm_base;
-	struct zwp_fullscreen_shell_v1 *fshell;
 	struct zwp_linux_dmabuf_v1 *dmabuf;
 	struct weston_direct_display_v1 *direct_display;
 	struct wp_viewporter *viewporter;
@@ -821,11 +819,6 @@ create_window(struct display *display, uint32_t win_flags)
 
 		window->wait_for_configure = true;
 		wl_surface_commit(window->surface);
-	} else if (display->fshell) {
-		zwp_fullscreen_shell_v1_present_surface(display->fshell,
-		                                        window->surface,
-		                                        ZWP_FULLSCREEN_SHELL_V1_PRESENT_METHOD_DEFAULT,
-		                                        NULL);
 	} else {
 		assert(0);
 	}
@@ -1130,10 +1123,6 @@ registry_handle_global(void *data, struct wl_registry *registry,
 		d->wm_base = wl_registry_bind(registry,
 					      id, &xdg_wm_base_interface, 1);
 		xdg_wm_base_add_listener(d->wm_base, &wm_base_listener, d);
-	} else if (strcmp(interface, zwp_fullscreen_shell_v1_interface.name) == 0) {
-		d->fshell = wl_registry_bind(registry,
-		                             id, &zwp_fullscreen_shell_v1_interface,
-		                             1);
 	} else if (strcmp(interface, zwp_linux_dmabuf_v1_interface.name) == 0) {
 		d->dmabuf = wl_registry_bind(registry,
 		                             id, &zwp_linux_dmabuf_v1_interface, 3);
@@ -1217,9 +1206,6 @@ destroy_display(struct display *display)
 	if (display->wm_base)
 		xdg_wm_base_destroy(display->wm_base);
 
-	if (display->fshell)
-		zwp_fullscreen_shell_v1_release(display->fshell);
-
 	if (display->compositor)
 		wl_compositor_destroy(display->compositor);
 
