@@ -48,6 +48,7 @@ struct kiosk_shell {
 
 	const struct weston_xwayland_surface_api *xwayland_surface_api;
 	struct weston_config *config;
+	struct wl_listener session_listener;
 };
 
 struct kiosk_shell_surface {
@@ -98,6 +99,8 @@ struct kiosk_shell_output {
 	struct wl_list link;
 
 	char *app_ids;
+	char *x11_wm_name_app_ids;
+	char *x11_wm_class_app_ids;
 
 	struct wl_list *active_surface_tree;
 };
