@@ -37,7 +37,13 @@
 
 #include "shared/helpers.h"
 #include "image-loader.h"
-#include "config-parser.h"
+#include "shared/xalloc.h"
+#include <libweston/config-parser.h>
+
+#ifdef HAVE_PANGO
+#include <fontconfig/fontconfig.h>
+#include <pango/pangocairo.h>
+#endif
 
 void
 surface_flush_device(cairo_surface_t *surface)
@@ -271,7 +277,6 @@ tile_source(cairo_t *cr, cairo_surface_t *surface,
 	pattern = cairo_pattern_create_for_surface (surface);
 	cairo_pattern_set_filter(pattern, CAIRO_FILTER_NEAREST);
 	cairo_set_source(cr, pattern);
-	cairo_pattern_destroy(pattern);
 
 	for (i = 0; i < 4; i++) {
 		fx = i & 1;
@@ -324,6 +329,9 @@ tile_source(cairo_t *cr, cairo_surface_t *surface,
 	cairo_rectangle(cr, x + width - margin, y + top_margin,
 			margin, height - margin - top_margin);
 	cairo_fill(cr);
+
+	cairo_pattern_destroy(pattern);
+	cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
 }
 
 void
@@ -340,25 +348,62 @@ rounded_rect(cairo_t *cr, int x0, int y0, int x1, int y1, int radius)
 	cairo_close_path(cr);
 }
 
+static void
+loaded_cairo_surface_destructor(void *data)
+{
+	struct weston_image *image = data;
+
+	weston_image_destroy(image);
+}
+
+static const cairo_user_data_key_t weston_cairo_util_load_cairo_surface_key;
+
 cairo_surface_t *
 load_cairo_surface(const char *filename)
 {
-	pixman_image_t *image;
+	cairo_surface_t *surface;
+	cairo_status_t ret;
+	struct weston_image *image;
 	int width, height, stride;
 	void *data;
 
-	image = load_image(filename);
+	image = weston_image_load(filename, WESTON_IMAGE_LOAD_IMAGE |
+					    WESTON_IMAGE_LOAD_ICC);
 	if (image == NULL) {
 		return NULL;
 	}
 
-	data = pixman_image_get_data(image);
-	width = pixman_image_get_width(image);
-	height = pixman_image_get_height(image);
-	stride = pixman_image_get_stride(image);
+	data = pixman_image_get_data(image->pixman_image);
+	width = pixman_image_get_width(image->pixman_image);
+	height = pixman_image_get_height(image->pixman_image);
+	stride = pixman_image_get_stride(image->pixman_image);
 
-	return cairo_image_surface_create_for_data(data, CAIRO_FORMAT_ARGB32,
-						   width, height, stride);
+	surface = cairo_image_surface_create_for_data(data, CAIRO_FORMAT_ARGB32,
+						      width, height, stride);
+	ret = cairo_surface_status(surface);
+	if (ret != CAIRO_STATUS_SUCCESS)
+		goto fail;
+
+	ret = cairo_surface_set_user_data(surface,
+					  &weston_cairo_util_load_cairo_surface_key,
+					  image,
+					  loaded_cairo_surface_destructor);
+	if (ret != CAIRO_STATUS_SUCCESS)
+		goto fail;
+
+	return surface;
+
+fail:
+	cairo_surface_destroy(surface);
+	weston_image_destroy(image);
+	return NULL;
+}
+
+struct weston_image *
+load_cairo_surface_get_user_data(cairo_surface_t *surface)
+{
+	return cairo_surface_get_user_data(surface,
+					   &weston_cairo_util_load_cairo_surface_key);
 }
 
 void
@@ -383,9 +428,7 @@ theme_create(void)
 	struct theme *t;
 	cairo_t *cr;
 
-	t = malloc(sizeof *t);
-	if (t == NULL)
-		return NULL;
+	t = xzalloc(sizeof *t);
 
 	t->margin = 32;
 	t->width = 6;
@@ -445,22 +488,67 @@ theme_create(void)
 void
 theme_destroy(struct theme *t)
 {
+#ifdef HAVE_PANGO
+	if (t->pango_context)
+		g_object_unref(t->pango_context);
+#endif
 	cairo_surface_destroy(t->active_frame);
 	cairo_surface_destroy(t->inactive_frame);
 	cairo_surface_destroy(t->shadow);
 	free(t);
 }
 
+#ifdef HAVE_PANGO
+static PangoLayout *
+create_layout(struct theme *t, cairo_t *cr, const char *title)
+{
+	PangoLayout *layout;
+	PangoFontDescription *desc;
+
+	if (!t->pango_context) {
+		PangoFontMap *fontmap;
+
+		fontmap = pango_cairo_font_map_new();
+		t->pango_context = pango_font_map_create_context(fontmap);
+		g_object_unref(fontmap);
+	}
+
+	pango_cairo_update_context(cr, t->pango_context);
+	layout = pango_layout_new(t->pango_context);
+	if (title) {
+		pango_layout_set_text(layout, title, -1);
+		desc = pango_font_description_from_string("sans-serif Bold 10");
+		pango_layout_set_font_description(layout, desc);
+		pango_font_description_free(desc);
+	}
+
+	pango_layout_set_ellipsize(layout, PANGO_ELLIPSIZE_END);
+	pango_layout_set_alignment(layout, PANGO_ALIGN_LEFT);
+	pango_layout_set_auto_dir (layout, FALSE);
+	pango_layout_set_single_paragraph_mode (layout, TRUE);
+	pango_layout_set_width (layout, -1);
+
+	return layout;
+}
+#endif
+
+#ifdef HAVE_PANGO
+#define SHOW_TEXT(cr) \
+	pango_cairo_show_layout(cr, title_layout)
+#else
+#define SHOW_TEXT(cr) \
+	cairo_show_text(cr, title)
+#endif
+
 void
 theme_render_frame(struct theme *t,
 		   cairo_t *cr, int width, int height,
-		   const char *title, struct wl_list *buttons,
-		   uint32_t flags)
+		   const char *title, cairo_rectangle_int_t *title_rect,
+		   struct wl_list *buttons, uint32_t flags)
 {
-	cairo_text_extents_t extents;
-	cairo_font_extents_t font_extents;
 	cairo_surface_t *source;
 	int x, y, margin, top_margin;
+	int text_width, text_height;
 
 	cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
 	cairo_set_source_rgba(cr, 0, 0, 0, 0);
@@ -491,36 +579,61 @@ theme_render_frame(struct theme *t,
 		    t->width, top_margin);
 
 	if (title || !wl_list_empty(buttons)) {
-		cairo_rectangle (cr, margin + t->width, margin,
-				 width - (margin + t->width) * 2,
-				 t->titlebar_height - t->width);
-		cairo_clip(cr);
 
+		cairo_rectangle (cr, title_rect->x, title_rect->y,
+				 title_rect->width, title_rect->height);
+		cairo_clip(cr);
 		cairo_set_operator(cr, CAIRO_OPERATOR_OVER);
-		cairo_select_font_face(cr, "sans",
+
+#ifdef HAVE_PANGO
+		PangoLayout *title_layout;
+		PangoRectangle logical;
+
+		title_layout = create_layout(t, cr, title);
+
+		pango_layout_get_pixel_extents (title_layout, NULL, &logical);
+		text_width = MIN(title_rect->width, logical.width);
+		text_height = logical.height;
+		if (text_width < logical.width)
+		  pango_layout_set_width (title_layout, text_width * PANGO_SCALE);
+
+#else
+		cairo_text_extents_t extents;
+		cairo_font_extents_t font_extents;
+
+		cairo_select_font_face(cr, "sans-serif",
 				       CAIRO_FONT_SLANT_NORMAL,
 				       CAIRO_FONT_WEIGHT_BOLD);
 		cairo_set_font_size(cr, 14);
 		cairo_text_extents(cr, title, &extents);
 		cairo_font_extents (cr, &font_extents);
-		x = (width - extents.width) / 2;
-		y = margin +
-			(t->titlebar_height -
-			 font_extents.ascent - font_extents.descent) / 2 +
-			font_extents.ascent;
+		text_width = extents.width;
+		text_height = font_extents.descent - font_extents.ascent;
+#endif
+
+		x = (width - text_width) / 2;
+		y = margin + (t->titlebar_height - text_height) / 2;
+		if (x < title_rect->x)
+			x = title_rect->x;
+		else if (x + text_width > (title_rect->x + title_rect->width))
+			x = (title_rect->x + title_rect->width) - text_width;
 
 		if (flags & THEME_FRAME_ACTIVE) {
 			cairo_move_to(cr, x + 1, y  + 1);
 			cairo_set_source_rgb(cr, 1, 1, 1);
-			cairo_show_text(cr, title);
+			SHOW_TEXT(cr);
 			cairo_move_to(cr, x, y);
 			cairo_set_source_rgb(cr, 0, 0, 0);
-			cairo_show_text(cr, title);
+			SHOW_TEXT(cr);
 		} else {
 			cairo_move_to(cr, x, y);
 			cairo_set_source_rgb(cr, 0.4, 0.4, 0.4);
-			cairo_show_text(cr, title);
+			SHOW_TEXT(cr);
 		}
+
+#ifdef HAVE_PANGO
+		g_object_unref(title_layout);
+#endif
 	}
 }
 
@@ -577,3 +690,23 @@ theme_get_location(struct theme *t, int x, int y,
 
 	return location;
 }
+
+/** Cleanup static Cairo/Pango data
+ *
+ * Using Cairo, Pango, PangoCairo, and fontconfig, ends up leaving a trail of
+ * thread-cached data behind us. Clean up what we can.
+ */
+void
+cleanup_after_cairo(void)
+{
+	/* some clients, particular weston-editor, still creates indirectly a
+	 * new font map; this makes sure we untie that up and avoid an assert
+	 * from cairo */
+#ifdef HAVE_PANGO
+	pango_cairo_font_map_set_default(NULL);
+#endif
+	cairo_debug_reset_static_data();
+#ifdef HAVE_PANGO
+	FcFini();
+#endif
+}
