@@ -32,6 +32,7 @@
 #include <assert.h>
 #include <errno.h>
 #include <unistd.h>
+#include <math.h>
 
 #include <libweston/config-parser.h>
 
@@ -47,23 +48,25 @@ static struct weston_config *
 load_config(const char *text)
 {
 	struct weston_config *config = NULL;
-	int len = 0;
-	int fd = -1;
-	char file[] = "/tmp/weston-config-parser-test-XXXXXX";
+	char *content = NULL;
+	size_t file_len = 0;
+	int write_len;
+	FILE *file;
 
-	ZUC_ASSERTG_NOT_NULL(text, out);
+	file = open_memstream(&content, &file_len);
+	ZUC_ASSERTG_NOT_NULL(file, out);
 
-	fd = mkstemp(file);
-	ZUC_ASSERTG_NE(-1, fd, out);
+	write_len = fwrite(text, 1, strlen(text), file);
+	ZUC_ASSERTG_EQ((int)strlen(text), write_len, out_close);
 
-	len = write(fd, text, strlen(text));
-	ZUC_ASSERTG_EQ((int)strlen(text), len, out_close);
+	ZUC_ASSERTG_EQ(fflush(file), 0, out_close);
+	fseek(file, 0L, SEEK_SET);
 
-	config = weston_config_parse(file);
+	config = weston_config_parse_fp(file);
 
 out_close:
-	close(fd);
-	unlink(file);
+	fclose(file);
+	free(content);
 out:
 	return config;
 }
@@ -119,6 +122,13 @@ static struct zuc_fixture config_test_t1 = {
 	"zero=0\n"
 	"negative=-42\n"
 	"flag=false\n"
+	"real=4.667\n"
+	"negreal=-3.2\n"
+	"expval=24.687E+15\n"
+	"negexpval=-3e-2\n"
+	"notanumber=nan\n"
+	"empty=\n"
+	"tiny=0.0000000000000000000000000000000000000063548\n"
 	"\n"
 	"[colors]\n"
 	"none=0x00000000\n"
@@ -322,54 +332,45 @@ ZUC_TEST_F(config_test_t1, test009, data)
 
 ZUC_TEST_F(config_test_t1, test010, data)
 {
-	int r, b;
+	int r;
+	bool b;
 	struct weston_config_section *section;
 	struct weston_config *config = data;
 
 	section = weston_config_get_section(config, "bar", NULL, NULL);
-	r = weston_config_section_get_bool(section, "flag", &b, 600);
+	r = weston_config_section_get_bool(section, "flag", &b, true);
 	ZUC_ASSERT_EQ(0, r);
-	ZUC_ASSERT_EQ(0, b);
+	ZUC_ASSERT_EQ(false, b);
 }
 
 ZUC_TEST_F(config_test_t1, test011, data)
 {
-	int r, b;
+	int r;
+	bool b;
 	struct weston_config_section *section;
 	struct weston_config *config = data;
 
 	section = weston_config_get_section(config, "stuff", NULL, NULL);
-	r = weston_config_section_get_bool(section, "flag", &b, -1);
+	r = weston_config_section_get_bool(section, "flag", &b, false);
 	ZUC_ASSERT_EQ(0, r);
-	ZUC_ASSERT_EQ(1, b);
+	ZUC_ASSERT_EQ(true, b);
 }
 
 ZUC_TEST_F(config_test_t1, test012, data)
 {
-	int r, b;
+	int r;
+	bool b;
 	struct weston_config_section *section;
 	struct weston_config *config = data;
 
 	section = weston_config_get_section(config, "stuff", NULL, NULL);
-	r = weston_config_section_get_bool(section, "flag", &b, -1);
-	ZUC_ASSERT_EQ(0, r);
-	ZUC_ASSERT_EQ(1, b);
+	r = weston_config_section_get_bool(section, "bonk", &b, false);
+	ZUC_ASSERT_EQ(-1, r);
+	ZUC_ASSERT_EQ(ENOENT, errno);
+	ZUC_ASSERT_EQ(false, b);
 }
 
 ZUC_TEST_F(config_test_t1, test013, data)
-{
-	int r, b;
-	struct weston_config_section *section;
-	struct weston_config *config = data;
-
-	section = weston_config_get_section(config, "stuff", NULL, NULL);
-	r = weston_config_section_get_bool(section, "bonk", &b, -1);
-	ZUC_ASSERT_EQ(-1, r);
-	ZUC_ASSERT_EQ(ENOENT, errno);
-	ZUC_ASSERT_EQ(-1, b);
-}
-
-ZUC_TEST_F(config_test_t1, test014, data)
 {
 	char *s;
 	int r;
@@ -387,7 +388,7 @@ out_free:
 	free(s);
 }
 
-ZUC_TEST_F(config_test_t1, test015, data)
+ZUC_TEST_F(config_test_t1, test014, data)
 {
 	char *s;
 	int r;
@@ -405,7 +406,7 @@ out_free:
 	free(s);
 }
 
-ZUC_TEST_F(config_test_t1, test016, data)
+ZUC_TEST_F(config_test_t1, test015, data)
 {
 	char *s;
 	int r;
@@ -425,7 +426,7 @@ out_free:
 	free(s);
 }
 
-ZUC_TEST_F(config_test_t1, test017, data)
+ZUC_TEST_F(config_test_t1, test016, data)
 {
 	const char *name;
 	int i;
@@ -440,7 +441,7 @@ ZUC_TEST_F(config_test_t1, test017, data)
 	ZUC_ASSERT_EQ(6, i);
 }
 
-ZUC_TEST_F(config_test_t1, test018, data)
+ZUC_TEST_F(config_test_t1, test017, data)
 {
 	int r;
 	int32_t n;
@@ -455,7 +456,7 @@ ZUC_TEST_F(config_test_t1, test018, data)
 	ZUC_ASSERT_EQ(0, errno);
 }
 
-ZUC_TEST_F(config_test_t1, test019, data)
+ZUC_TEST_F(config_test_t1, test018, data)
 {
 	int r;
 	uint32_t n;
@@ -470,7 +471,7 @@ ZUC_TEST_F(config_test_t1, test019, data)
 	ZUC_ASSERT_EQ(0, errno);
 }
 
-ZUC_TEST_F(config_test_t1, test020, data)
+ZUC_TEST_F(config_test_t1, test019, data)
 {
 	int r;
 	uint32_t n;
@@ -485,7 +486,7 @@ ZUC_TEST_F(config_test_t1, test020, data)
 	ZUC_ASSERT_EQ(0, errno);
 }
 
-ZUC_TEST_F(config_test_t1, test021, data)
+ZUC_TEST_F(config_test_t1, test020, data)
 {
 	int r;
 	uint32_t n;
@@ -500,7 +501,7 @@ ZUC_TEST_F(config_test_t1, test021, data)
 	ZUC_ASSERT_EQ(0, errno);
 }
 
-ZUC_TEST_F(config_test_t1, test022, data)
+ZUC_TEST_F(config_test_t1, test021, data)
 {
 	int r;
 	uint32_t n;
@@ -515,7 +516,7 @@ ZUC_TEST_F(config_test_t1, test022, data)
 	ZUC_ASSERT_EQ(0, errno);
 }
 
-ZUC_TEST_F(config_test_t1, test023, data)
+ZUC_TEST_F(config_test_t1, test022, data)
 {
 	int r;
 	uint32_t n;
@@ -531,7 +532,7 @@ ZUC_TEST_F(config_test_t1, test023, data)
 	ZUC_ASSERT_EQ(0, errno);
 }
 
-ZUC_TEST_F(config_test_t1, test024, data)
+ZUC_TEST_F(config_test_t1, test023, data)
 {
 	int r;
 	uint32_t n;
@@ -547,7 +548,7 @@ ZUC_TEST_F(config_test_t1, test024, data)
 	ZUC_ASSERT_EQ(0, errno);
 }
 
-ZUC_TEST_F(config_test_t1, test025, data)
+ZUC_TEST_F(config_test_t1, test024, data)
 {
 	int r;
 	uint32_t n;
@@ -563,7 +564,7 @@ ZUC_TEST_F(config_test_t1, test025, data)
 	ZUC_ASSERT_EQ(EINVAL, errno);
 }
 
-ZUC_TEST_F(config_test_t1, test026, data)
+ZUC_TEST_F(config_test_t1, test025, data)
 {
 	int r;
 	uint32_t n;
@@ -579,7 +580,7 @@ ZUC_TEST_F(config_test_t1, test026, data)
 	ZUC_ASSERT_EQ(EINVAL, errno);
 }
 
-ZUC_TEST_F(config_test_t1, test027, data)
+ZUC_TEST_F(config_test_t1, test026, data)
 {
 	int r;
 	int32_t n;
@@ -594,7 +595,7 @@ ZUC_TEST_F(config_test_t1, test027, data)
 	ZUC_ASSERT_EQ(0, errno);
 }
 
-ZUC_TEST_F(config_test_t1, test028, data)
+ZUC_TEST_F(config_test_t1, test027, data)
 {
 	int r;
 	uint32_t n;
@@ -609,6 +610,197 @@ ZUC_TEST_F(config_test_t1, test028, data)
 	ZUC_ASSERT_EQ(ERANGE, errno);
 }
 
+ZUC_TEST_F(config_test_t1, get_double_number, data)
+{
+	int r;
+	double n;
+	struct weston_config_section *section;
+	struct weston_config *config = data;
+
+	errno = 0;
+	section = weston_config_get_section(config, "bar", NULL, NULL);
+	r = weston_config_section_get_double(section, "number", &n, 600.0);
+
+	ZUC_ASSERT_EQ(0, r);
+	ZUC_ASSERT_TRUE(5252.0 == n);
+	ZUC_ASSERT_EQ(0, errno);
+}
+
+ZUC_TEST_F(config_test_t1, get_double_missing, data)
+{
+	int r;
+	double n;
+	struct weston_config_section *section;
+	struct weston_config *config = data;
+
+	section = weston_config_get_section(config, "bar", NULL, NULL);
+	r = weston_config_section_get_double(section, "+++", &n, 600.0);
+
+	ZUC_ASSERT_EQ(-1, r);
+	ZUC_ASSERT_TRUE(600.0 == n);
+	ZUC_ASSERT_EQ(ENOENT, errno);
+}
+
+ZUC_TEST_F(config_test_t1, get_double_zero, data)
+{
+	int r;
+	double n;
+	struct weston_config_section *section;
+	struct weston_config *config = data;
+
+	errno = 0;
+	section = weston_config_get_section(config, "bar", NULL, NULL);
+	r = weston_config_section_get_double(section, "zero", &n, 600.0);
+
+	ZUC_ASSERT_EQ(0, r);
+	ZUC_ASSERT_TRUE(0.0 == n);
+	ZUC_ASSERT_EQ(0, errno);
+}
+
+ZUC_TEST_F(config_test_t1, get_double_negative, data)
+{
+	int r;
+	double n;
+	struct weston_config_section *section;
+	struct weston_config *config = data;
+
+	errno = 0;
+	section = weston_config_get_section(config, "bar", NULL, NULL);
+	r = weston_config_section_get_double(section, "negative", &n, 600.0);
+
+	ZUC_ASSERT_EQ(0, r);
+	ZUC_ASSERT_TRUE(-42.0 == n);
+	ZUC_ASSERT_EQ(0, errno);
+}
+
+ZUC_TEST_F(config_test_t1, get_double_flag, data)
+{
+	int r;
+	double n;
+	struct weston_config_section *section;
+	struct weston_config *config = data;
+
+	errno = 0;
+	section = weston_config_get_section(config, "bar", NULL, NULL);
+	r = weston_config_section_get_double(section, "flag", &n, 600.0);
+
+	ZUC_ASSERT_EQ(-1, r);
+	ZUC_ASSERT_TRUE(600.0 == n);
+	ZUC_ASSERT_EQ(EINVAL, errno);
+}
+
+ZUC_TEST_F(config_test_t1, get_double_real, data)
+{
+	int r;
+	double n;
+	struct weston_config_section *section;
+	struct weston_config *config = data;
+
+	errno = 0;
+	section = weston_config_get_section(config, "bar", NULL, NULL);
+	r = weston_config_section_get_double(section, "real", &n, 600.0);
+
+	ZUC_ASSERT_EQ(0, r);
+	ZUC_ASSERT_TRUE(4.667 == n);
+	ZUC_ASSERT_EQ(0, errno);
+}
+
+ZUC_TEST_F(config_test_t1, get_double_negreal, data)
+{
+	int r;
+	double n;
+	struct weston_config_section *section;
+	struct weston_config *config = data;
+
+	errno = 0;
+	section = weston_config_get_section(config, "bar", NULL, NULL);
+	r = weston_config_section_get_double(section, "negreal", &n, 600.0);
+
+	ZUC_ASSERT_EQ(0, r);
+	ZUC_ASSERT_TRUE(-3.2 == n);
+	ZUC_ASSERT_EQ(0, errno);
+}
+
+ZUC_TEST_F(config_test_t1, get_double_expval, data)
+{
+	int r;
+	double n;
+	struct weston_config_section *section;
+	struct weston_config *config = data;
+
+	errno = 0;
+	section = weston_config_get_section(config, "bar", NULL, NULL);
+	r = weston_config_section_get_double(section, "expval", &n, 600.0);
+
+	ZUC_ASSERT_EQ(0, r);
+	ZUC_ASSERT_TRUE(24.687e+15 == n);
+	ZUC_ASSERT_EQ(0, errno);
+}
+
+ZUC_TEST_F(config_test_t1, get_double_negexpval, data)
+{
+	int r;
+	double n;
+	struct weston_config_section *section;
+	struct weston_config *config = data;
+
+	errno = 0;
+	section = weston_config_get_section(config, "bar", NULL, NULL);
+	r = weston_config_section_get_double(section, "negexpval", &n, 600.0);
+
+	ZUC_ASSERT_EQ(0, r);
+	ZUC_ASSERT_TRUE(-3e-2 == n);
+	ZUC_ASSERT_EQ(0, errno);
+}
+
+ZUC_TEST_F(config_test_t1, get_double_notanumber, data)
+{
+	int r;
+	double n;
+	struct weston_config_section *section;
+	struct weston_config *config = data;
+
+	errno = 0;
+	section = weston_config_get_section(config, "bar", NULL, NULL);
+	r = weston_config_section_get_double(section, "notanumber", &n, 600.0);
+
+	ZUC_ASSERT_EQ(0, r);
+	ZUC_ASSERT_TRUE(isnan(n));
+	ZUC_ASSERT_EQ(0, errno);
+}
+
+ZUC_TEST_F(config_test_t1, get_double_empty, data)
+{
+	int r;
+	double n;
+	struct weston_config_section *section;
+	struct weston_config *config = data;
+
+	errno = 0;
+	section = weston_config_get_section(config, "bar", NULL, NULL);
+	r = weston_config_section_get_double(section, "empty", &n, 600.0);
+
+	ZUC_ASSERT_EQ(0, r);
+	ZUC_ASSERT_TRUE(0.0 == n);
+	ZUC_ASSERT_EQ(0, errno);
+}
+
+ZUC_TEST_F(config_test_t1, get_double_tiny, data)
+{
+	int r;
+	double n;
+	struct weston_config_section *section;
+	struct weston_config *config = data;
+
+	errno = 0;
+	section = weston_config_get_section(config, "bar", NULL, NULL);
+	r = weston_config_section_get_double(section, "tiny", &n, 600.0);
+
+	ZUC_ASSERT_EQ(0, r);
+	ZUC_ASSERT_TRUE(6.3548e-39 == n);
+	ZUC_ASSERT_EQ(0, errno);
+}
+
 ZUC_TEST_F(config_test_t2, doesnt_parse, data)
 {
 	struct weston_config *config = data;
